[include version.inc]
[comment {-*- tcl -*- doctools manpage}]
[manpage_begin "tarray_guide" n [vset tarray_version]]
[copyright "2012-2013 Ashok P. Nadkarni"]
[moddesc "Typed Arrays"]
[titledesc "Typed Arrays Programmer's Guide"]

[description]
This document is a programmer's guide for installing and using the
[cmd tarray] extension (Version [vset tarray_version]) from Tcl.

[section "Installation and Loading"]
Binary packages for some platforms are available from the
[uri http://www.sourceforge.net/projects/tarray "SourceForge Tarray download"]
area. See [uri build.html "Building tarray"] for instructions to
build the extension for other platforms.
[para]
To install the extension,
extract the files from the distribution to any directory that is included
in your Tcl installation's [var auto_path] variable.
[para]
Once installed, the extension can be loaded with the standard Tcl
[cmd "package require"] command.
[example "package require tarray"]
[section Concepts]

[subsection "Columns and Tables"]
The extension implements two data types - [uri column.html columns]
and [uri table.html tables]. The general term [emph "typed array"]
is used to refer to either of these.

A typed column is an array containing elements of a single 
[sectref "Types" type]
that is specified when the column is created. The command 
[uri column.html [cmd tarray::column]] can be used
to create and manipulate typed columns.
operations.
[para]
A typed table is an ordered sequence of named columns
of equal size. It can be also be viewed as an array of records where the
record fields happen to use column-wise storage. The corresponding
[uri table.html [cmd tarray::table]] command operates on typed tables.
Columns in a table can be referenced using either their name or
their position in the ordered sequence.

[include types.inc]

[include indices.inc]

[section "Using Typed Arrays"]

The [cmd tarray] extension implements two commands,
[uri column.html [cmd column]] and [uri table.html [cmd table]],
both in the [cmd tarray::] namespace. Each is an ensemble of subcommands
that operate on columns and table respectively.
The examples below assume the commands have been imported
into the calling namespace or are in its namespace path.

[subsection "Creating Typed Arrays"]

The [cmd create] subcommand creates columns and tables.

[example {column create int}]

will create a typed column that can hold element of the [cmd int] type.
The column can be initialized at creation time.

[example {column create int {0 1 2 3}}]

initializes the first four elements of the array. 

[para]
The array will be grown as needed but as an optimization, preallocation
may be requested.

[example {column create int {0 1 2 3} 1000}]

will request a preallocation of a thousand elements with the first four
being initialized.

Tables can be initialized in analogous fashion, for example, to create
a initialized table

[example_begin]
set tab [lb]table create {country string population wide} {{China 1350000000} {Vatican 850}}[rb]
[example_end]

[subsection "Specifying Indices"]

Most commands require specification of the array locations
to be targeted. This specification can be
[list_begin itemized]
[item] a single index,
[item] a contiguous range of indices, or 
[item] a list of indexes in arbitrary order specified as a list of integers
or a column of type integer.
[list_end]
[example_begin]
set col [lb]column create double {}[rb]
set col [lb]column fill $col 1.0 0 9[rb]
set col [lb]column fill $col 2.0 3[rb]
set col [lb]column fill $col 2.0 end-2 end[rb]
set col [lb]column fill $col 3.0 {2 7}[rb]
set col [lb]column fill $col 3.0 [lb]column create int {2 7}[rb][rb]
[example_end]
The initial [cmd fill] initializes the first 10 elements of the column
to [const 1.0]. The second one sets the single element at index [const 3]
to [const 2.0]. The sets the last 3 elements to [const 2.0]. The last two
set elements at indices [const 2] and [const 7] to [const 3.0].
[para]
The last form, an integer column, is useful because some commands return
indices in that form. For example, the following will replace all elements
with negative values with [const 0.0].
[example_begin]
set col [lb]column fill $col 0.0 [lb]column search -all -lt $col 0]]
[example_end]

[para]
Although the above example used columns, table indices are specified
in identical fashion.

[subsection "Values and Variables"]
Commands that modify typed arrays come in two flavors:
[list_begin itemized]
[item] Commands that operate on
column and table values and return the modified column or table as a result
(for example [cmd fill]), and
[item] Commands that modify a Tcl [emph variable] containing the column or table
(for example [cmd vfill]).
[list_end]
The difference is similar to how different
Tcl list commands behave, e.g. [cmd linsert] and [cmd lreplace] versus
[cmd lset] and [cmd lappend].

[para]
The examples above used the value-oriented form of the commands where
the [cmd fill] modifies a copy of the contents of the [variable col]
and returns the modified copy which is then stored back into [variable col].
For large typed arrays, this is inefficient and the above would be better
written as
[example_begin]
set col [lb]column create double {}]
column vfill col 1.0 0 9]
column vfill col 2.0 3]
column vfill col 3.0 {2 7}]
column vfill col 3.0 [lb]column create int {2 7}]]
[example_end]
Here the [cmd vfill] command is directly modifying the variable [variable col]
and assuming the content is not shared, no copy needs to be made.

[para]
Almost every command that modifies a typed array has this dual equivalent.

[comment "TBD - expand on the efficiency issues and benefits"]
[comment "TBD - expand on the string versus any column type"]

[subsection "Modifying Data"]

Modifying a typed array may involve either storing a 
[emph single]
value at multiple target locations or a different value at each target location.
Further, the locations may be a contiguous range or a noncontiguous set of
indices.
[list_begin itemized]
[item] The [cmd fill/vfill] commands store a [emph single] value at one or more
locations, either contiguous or noncontiguous.
[item] The [cmd place/vplace] commands store each value from a 
sequence of values at one or more [emph noncontiguous] locations.
[item] The [cmd put/vput] commands store each value from a sequence of 
values in [emph contiguous] locations starting at a specified index.
[list_end]
The sequence of values to be stored may be specified as a Tcl list or
a typed array. When multiple noncontiguous target locations are specified,
they may be specified as a Tcl list of integers or an [const int] column.

[example_begin]
column vplace col {300.0 200.0 500.0} {3 2 5}
column vput col {7.0 8.0 9.0} 6
column vput col {11.0 12.0}
[example_end]
The first command stores values [const 300.0], [const 200.0] and
[const 500.0] at indices [const 3], [const 2] and [const 5]. Note the indices
do not have to be specified in order. The second command stores
[const 7.0], [const 8.0] and [const 9.0] in successive locations starting
at index [const 6]. The last command appends [const 11.0], [const 12.0]
to the end of the column.
[para]
Instead of specifying values as a list, they may also be specified
as a column of the same type.
[example_begin]
set colA [lb]column create double {1.0 2.0 4.0}]
column vplace colB $colA {3 2 5}
[example_end]
Again, this form is particularly useful when storing columns returned
from commands into another column.

[para]
The [cmd table] command has equivalent commands. For example
[example_begin]
set populations [lb]table create {country string population wide} {}[rb]
table vput populations {{China 1350000000} {Vatican 850}}
[example_end]
Note that just like in the case of columns, the list of values can
be specified as a table instead, provided the column types are the same.
[para]
When storing data, the [cmd -columns] option comes in handy for two
different purposes. First, it allows data to be specified in a different
order than that specified in the column definition. For example,
in the above example, if the order of the supplied data was population
followed by country, the command could have been written as follows:
[example_begin]
table vput -columns {population country} populations {{1350000000 China} {850 Vatican}}
[example_end]
There is no need to reorder the fields in the input data.
[para]
Secondly, the [cmd -columns] allows modification of a subset of the columns.
[example_begin]
set populations [lb]table create {country string population wide} {}[rb]
table vput populations {{China 1350000000} {Vatican 850}}
table vfill -columns {population} populations {851} 1
[example_end]
The population of the second table row is changed to 851.

[para]
A point to be noted about all the above commands is that they may
extend the size of the array if necessary. However, two conditions
must apply for this: 
[list_begin itemized]
[item]
where an index list or index column is specified, there must not
be any gaps in indices that extend the array.
[example_begin]
set populations [lb]table create {country string population wide} {{Vatican 850}  {China 1350000000}}[rb]
table put populations {{Vatican 860} {India 1250000000} {USA 314000000}} {0 3 2}
table put populations {{Vatican 860} {India 1250000000} {USA 314000000}} {1 3 4}
[example_end]
The first [cmd put] will succeed, changing the existing value at index 1
and extending the array by two rows (note order of indices does not matter).
The second [cmd put] will raise an error since index [const 2] neither exists
nor is supplied in the command.
[item]
Second, if the [cmd -columns] option is specified, it must include all
columns (in any order) of the table. Otherwise, the command will not
know what value to use for the other columns when extending the table.
[list_end]

[para]
All the commands dicussed to this point overwrite an existing values, at
the target locations. The [cmd "column insert"] and [cmd "column vinsert"]
commands and their table equivalents store one value, or multiple values
(passed as a list or a typed array),  into
the type array, pushing existing elements further up.
[example_begin]
column insert $col 3.0 2 10]
column insert $col {1.0 2.0 3.0} 2
column insert $col $col 2
[example_end]
The first command returns a new column with the same value, [const 3.0],
inserted [const 10] times at index [const 2]. The second command returns
a new column with 
all values in the passed list, [const 1.0], [const 2.0], [const 3.0],
inserted at index [const 2].
The last command returns a new column where all existing values in the
column are reinserted at index [const 2].

[para]
Elements in a typed array can be deleted with the [cmd delete/vdelete]
commands. Succeeding elements are moved up to occupy the deleted slots.
Like the [cmd fill] command, the indices of the elements
to be deleted may be specified as a single index, a range, a list of
indices or a index column.
[example_begin]
column vdelete col [lb]column search -all -lt $col 0]
[example_end]
will delete all negative elements from the column stored in [variable col].

[para]
Finally, typed arrays support [cmd reverse/vreverse] commands
which reverse the order of elements, an operation that is useful in
many algorithms.

[subsection "Retrieving Data"]

Retrieving data from a typed array involves specifying which elements 
to retrieve and what format to retrieve them in when multiple elements
are retrieved.

[para]
As usual, the elements to be retrieved can be specified as a single index,
an index range, a list of indices or an index column. In the simplest
cast, the [cmd index] command can be used to retrieve a single element.
[example_begin]
column index $col 4
table index $tab end
[example_end]

Multiple elements can be retrieved with the [cmd get]
and [cmd range] commands. The [cmd get] command
can be passed a sequence of noncontiguous indices specified
as a Tcl list or a [const int] column:
[example_begin]
column get $col {10 7 4}
table get $tab [lb]column search -all -lt [lb]table column $tab 0] 0]
[example_end]

The [cmd range] command retrieves elements in a specified index range.
[example_begin]
column range $col 3 5
table range $tab 0 10
[example_end]

[para]
By default, both commands returns values as a typed array.
The [cmd -list] and [cmd -dict] commands can be used to return the values
as a Tcl list or dictionary instead. In the latter case, the dictionary
keys are the indices being retrieved.
[example_begin]
column get -list $col 3 5
table get -dict $tab [lb]column search -all -lt [lb]table column $tab 0] 0]
[example_end]

[para]
In the case of tables, both commands also provide for retrieval of
a subset of columns and in a different order than in the definition.
[example_begin]
table range -columns {population country} $populations 0 end
table range -columns {1} $populations 0 end
[example_end]
Note columns may be specified either by position or name.

[para]
Tables provide additional commands for retrieving entire columns.
[list_begin itemized]
[item] [cmd "table column"] returns a column from a table. This is
useful for sorting and searching columns as shown in table search
examples below.
[item] [cmd "table slice"] returns a new table containing a subset
of the columns of a table.
[list_end]

[subsection "Searching"]
The [cmd "column search"] command works similarly to Tcl's [cmd lsearch].
It returns the indices (by default) or the values 
(with the [cmd -inline] option) of matching elements in a column.
Like [cmd lsearch], [cmd "column search"] stops on the first match and returns
the matching index or value but
the [cmd -all] option can be used to return all matches. The command supports
several matching operators. See the [uri column.html "column reference"]
for a list.
[para]
[example {column search $col 0}]
returns the index of the first element that is 0 using the default
matching operator that tests for equality
(assumes [variable col] is a numeric column).
[example {column search -inline -gt $col 0}]
returns the value of the first positive element.
[example {column search -all -gt $col 0}]
returns the indices of all positive elements. The return value is
an [const int] column.
[example {column search -all -inline -nocase -pat $col *.exe}]
returns the values of all elements that match [const *.exe] using
case-insensitive matching as in Tcl's [cmd "string match -nocase"].

[para]
The search can be restricted to only look at specific elements using
a combination of [cmd -range] and [cmd -among] options.
[example {column search -range {0 9} $col 0}]
limits the search to the first ten elements.
[example {column search -among {1 5 3} $col 0}]
only examines the elements at positions [const 1], [const 5],
and [const 3] [emph "in that order"]. The option [cmd -among] is 
particularly useful in combining searches as in the table example below.

[para]
To search tables, use the [cmd search] on individual columns. For example,
[example_begin]
set countries [lb]table create {country string population wide area double} {
    {Vatican      850     0.44}
    {China 1350000000     9.55e6}
    {USA    314000000     9.63e6}
    {India 1250000000     3.3e6}
    {Russia 141930000     17e6}
}[rb]
set pop_col [lb]table column $countries 1]; # Column specified by position
set area_col [lb]table column $countries area]; # Column specified by name
table get -list -columns {country} $countries \
    [lb]column search -all \
        -among [lb]column search -all -gt $pop_col 250000000] \
        -gt $area_col 5e6]
[example_end]
returns names of countries that are populous and large in area. Note
how the outside search is limited to specific indices using the [cmd -among]
option.
[para]
The [cmd intersect3] command offers another way to search across multiple
columns as described later.

[subsection "Sorting"]
Columns can be sorted using the [cmd "column sort"] command or its
variable targeting analogue [cmd "column vsort"].
The commands take the [cmd -increasing] and [cmd -decreasing] options
to determine the sort order.
[para]
The [cmd "column sort"] command also takes
the [cmd -indices] option which results in the indices being returned 
instead of the values themselves. This is useful for sorting tables
based on a column. For example, assuming variable [var countries] has been
initialized as above,
[example_begin]
table get -list $countries [lb]column sort -indices -nocase [lb]table column $countries 0]]
[example_end]
returns rows in the sorted order based on country name.
[para]
When sorting tables, for display purposes for example, it is often
necessary to display elements that have the same value in the sort column
in the same order that they were previously displayed. Although, individual
column sorts are stable, this is not enough when sorting across multiple
columns. In such cases, the [cmd -indirect] option to the sort command
provides a solution. Using this option allows sorting where the
"initial" ordering of elements is different from the actual order
of elements in the column. An example will clarify this.
[para]
Consider a table that stores heights and weights.
[example_begin]
set tab [lb]table create {name string height int weight int} {
    {Jeff 180 80}
    {John 175 80}
    {Jim 170 75}
}[rb]
[example_end]
The user may choose to sort the table by height which boils down to
the following code:
[example_begin]
table get -list $tab [lb]column sort -indices [lb]table column $tab height[rb][rb]
[example_end]
This results in the table being displayed in the order
[const Jim], [const John], [const Jeff].
The user may then choose to sort by weight.
[example_begin]
table get -list $tab [lb]column sort -indices [lb]table column $tab weight[rb][rb]
[example_end]
resulting in a display in order
[const Jim], [const Jeff], [const John]. 
Since they actually have the same value in the new sort column,
this interchange of positions
between [const Jeff] and [const John] is disconcerting to the user.
Use of the [cmd -indirect] option overcomes this problem.
[example_begin]
set indices [lb]column sort -indices [lb]table column $tab height[rb][rb]
table get -list $tab $indices
# Now use previous order of indices to order elements when their
# values in the weight column are equal
table get -list $tab [lb]column sort -indirect [lb]table column weight] $indices[rb]
[example_end]
In this last statement, the sort is done indirectly using values
from table [variable tab] but the positioning of elements when
these values compare equal is based on the order in [variable indices].

[subsection "Other Commands"]

The [cmd "column count"] can be used to count the number values
in a column that match specified criteria.
[para]
The [cmd "column lookup"] provides for a faster, dictionary based
retrieval for string columns. This may be beneficial for columns used
as keys in a table.
[para]
The [cmd "column minmax"] command returns a pair containing the
minimum and maximum values in a column.
[para]
The [cmd "column intersect3"] command returns the differences between
two columns. The multiple column search example above could also
have written as follows:

[example_begin]
set pop_col [lb]table column $countries population]
set area_col [lb]table column $countries area]
set populous [lb]column search -all -gt $pop_col 250000000[rb]
set large [lb]column search -all -gt $area_col 5e6[rb]
table get -list $countries [lb]lindex [lb]column intersect $populous $large[rb] 0[rb]
[example_end]
In most cases, the previous method using [cmd "column search -among"]
is likely to be faster. However, the [cmd "column intersect3"] may be
faster in some cases, for example, when multiple combinations are desired.
[example_begin]
lassign [lb]column intersect $populous $large[rb] populous_and_large populous_but_small sparse_but_large
[example_end]

[para]
The [cmd "column size"] command returns the number of elements in a column.
The [cmd "table size"] and [cmd "table width"] commands return the number
of rows and columns respectively in a table.

[para]
The [cmd "column type"] command returns the type of elements in a column
while the [cmd "table types"] command returns a list of the types of each
column in the table.

[subsection "Usage Hints"]
[list_begin itemized]

[item]
When specifying indices to commands, Tcl lists of integers and 
columns of type [const int] are usually interchangeable. Similarly, when
passing multiple values to a command, either a Tcl list or a column
of the appropriate type can be used. Note there is an ambiguity in the
specific case where the target of the command is a column of type
[const any] and the passed operand is a string of the form
[example {tarray any {...}}]
where the operand can be interpreted either as a column or a Tcl 
list with three elements [const tarray], [const any] and the [const "{...}"].
In this case the operand gets interpreted as a column.

[item]
Given that the [package tarray] extension is meant for dealing with large
amounts of data, it is useful to keep in mind Tcl's object reference counting
and copy-on-write implementation. Modifying a typed array that is shared
will result in a copy being made, which can be expensive if the array is large.
So, to modify a variable that contains a typed array, the command
[example {column vput col [list 1.0 2.0]}]
is far more efficient than
[example {set col [column put $col [list 1.0 2.0]]}]
assuming the value in [variable col] is not itself shared. This is similar
to use of Tcl's [cmd lset] command to modify lists.

[item]
As arrays get large [package tarray] prioritizes memory usage over
efficiency. As arrays grow, the additional extra memory is conservatively
allocated (unlike Tcl which aggressively allocates extra memory).
If the size of a typed array can be estimated in advance, for example,
reading records from a database, the memory can be preallocated.
[example {column create int {} 1000000}]
preallocates space for a million elements.

[item]
Typed arrays are by design implemented as consecutive elements in
contiguous memory. Certain operations, such as insertion and deletion,
will not be efficient when arrays get very large. For applications
where such opertions are common, other structures should be built
on top using typed arrays as the lower level building blocks. Such
higher level structures can be scripted and customized for specific
usage patterns easily as they can be implemented at the script level
using the low level typed array operations for efficiency. Whether
this is required or not should be determined based on application
benchmarks.

[item]
Both lists and columns have some differences in terms of functionality.
Columns do not have the [cmd -stride] option but the same functionality
can be implemented through tables. List indexing offers nesting while
although columns can be nested, the nested columns have to be explicitly
accessed. On the other hand, columns offer some additional functions such 
as [cmd intersect3] and indexing operations 
(eg. extraction or storing of multiple elements through index lists).

[item]
Columns of type [const any] are stored as Tcl_Obj objects internally
and thus are very similar to Tcl lists.
Any advantage of an [const any] typed array over using a simple 
Tcl list in terms of the memory footprint comes only from conservative
memory overallocation, not from reduced memory size of individual elements.
It is therefore not as big a benefit as for other types.
Thus columns of type [const any] are mostly beneficial when used
in conjunction with other column types, for example in a table.

[item]
The type [const any] can be any Tcl value, including typed arrays. Typed
arrays can therefore be nested 
(tables are currently implemented as nested columns).
However, unlike some of the Tcl list commands, [package tarray] does
not have commands that implicitly support nesting. Nested
typed arrays have to be explicitly accessed as such.
[example {column index [column index $outer_column 4] 0}]

[item]
The package internally keeps track of the sorting state of a column.
A column is internally marked after certain operations where the result
is known to be sorted. An obvious example is the [cmd "column sort"]
command. A less obvious case is an index column returned from certain
search operations. Several commands make use of this for more efficient
operation. For example, the [cmd "column intersect3"] command is much
faster when columns are known to be sorted. Thus finding the intersection
of two index columns resulting from searches is an O(n) operation.

[list_end]

[manpage_end]
