[include version.inc]
[comment {-*- tcl -*- doctools manpage}]
[manpage_begin "tarray::table" n [vset tarray_version]]
[copyright "2012-2013 Ashok P. Nadkarni"]
[moddesc "Typed Array Tables"]
[titledesc "Commands related to typed array tables"]
[require Tcl 8.5]
[require tarray ?[vset tarray_version]?]

[description]

The [cmd tarray] extension implements typed arrays and associated commands
[uri column.html [cmd column]] and [cmd table].
This page provides reference documentation for
commands related to typed tables.
See [uri tarray.html] for an overview and [uri guide.html] for a
programming guide.
[subsection Tables]
A typed table is an ordered sequence of [uri tarray.html "typed columns"]
of equal size. It can be viewed as an array of records where the
record fields happen to use column-wise storage. The corresponding
[cmd table] command operates on typed tables.
[para]
The columns in a table are defined with a name, type and order when the table
is created. Commands that operate on tables allow columns to be specified
using either the column name or its position.

[include types.inc]
[include indices.inc]

[subsection "Standard Options"]
Many commands take one or more of the standard options described below.
The [cmd -list], [cmd -dict] and [cmd -table] options control
the format of the returned values. The [cmd -columns] option allows
selection and ordering of specific columns from the table.
[list_begin options]
[opt_def -columns [arg COLUMNS]]
Selects a subset of columns from the table and their order. If this option
is not specified, all columns in the table are selected as the target
of the command and in the same order as in the table definition.
If this option is specified, [arg COLUMNS] is a list of column indexes
and names. For commands that retrieve data, like [cmd "table get"],
only data from the specified columns is retrieved and in the column
order specified in [arg COLUMNS]. For commands that modify data,
only data in the specified columns is modified. The input data
values are taken in the same order as specified in [arg COLUMNS].
Note that when a command invocation that causes a table to grow specifies
the [cmd -columns] option, all columns must be included in [arg COLUMNS]
although they may be specified in any order depending on the order of
the source data.
[opt_def -dict]
Specifies that the return values must be in the form of a dictionary
keyed by the corresponding indices. The value of each key is a row
which is a list each element
of which is the value at that index in the corresponding column.
[opt_def -list] Specifies that the return value must be in the form
of a list of rows, one per index. Each row is itself a list, each element
of which is the value at that index in the corresponding column.
[opt_def -table]
Specifies that the values are to be returned as a table.
This is the default if neither [cmd -list] nor [cmd -dict] is specified.
[list_end]


[section "Command Reference"]

Note: All commands are located in the [cmd tarray] namespace.

[list_begin definitions]

[call [cmd "table column"] [arg TABLE] [arg COLSPEC]]
Returns a typed column from a table. [arg COLSPEC] may be either
the column name or its position.

[call [cmd "table cnames"] [arg TABLE]]
Returns the list of column names for the table.

[call [cmd "table create"] [arg DEFINITION] [arg ROWVALUES]]
Returns a typed table containing a sequence of columns.
[arg DEFINITION] is a list of alternating column names and column types.
A column name is an identifier for a column that can be used in lieu
of a column index. The type for a column must be one
of the valid types described in [sectref "Types"].
[para]
[arg ROWVALUES] is the initial content of the table array specified
as a nested list with each sublist being a row
whose element types are compatible with the corresponding
column types in [arg DEFINITION].

[call [cmd "table delete"] [arg TABLE] [arg INDEX]]
[call [cmd "table delete"] [arg TABLE] [arg INDEXLIST]]
[call [cmd "table delete"] [arg TABLE] [arg LOW] [arg HIGH]]
Returns a typed table with all specified rows deleted. The row
indices are specified in any of the forms described in [sectref Indices].

[call [cmd "table fill"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLE] [arg ROWVALUE] [arg INDEX]]
[call [cmd "table fill"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLE] [arg ROWVALUE] [arg INDEXLIST]]
[call [cmd "table fill"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLE] [arg ROWVALUE] [arg LOW] [arg HIGH]]
Returns a typed table with specified rows set to [arg ROWVALUE]. 
Each element from the list [arg ROWVALUE] is assigned to the corresponding
column of the table at the specified indices. Any additional elements
in [arg ROWVALUE] are ignored. An error is generated if any element of
[arg ROWVALUE] does not match the type of the corresponding column or
if [arg ROWVALUE] contains fewer elements than the table width.
Indices are specified in any of the forms described in [sectref Indices].
The size of the array will be extended if necessary.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to target specific columns of the table or to match
the order of columns to the supplied data.

[call [cmd "table get"] [arg [opt "OPTIONS"]] [arg TABLE] [arg INDEXLIST]]
Returns the values from a table at the indices specified as
[sectref Indices "index list"].
Any of the [sectref "Standard Options" "standard options"]
may be specified with this command.

[call [cmd "table index"] [arg TABLE] [arg INDEX]]
Returns the value of the row at the position specified by [arg INDEX].

[call [cmd "table insert"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLE] [arg ROWVALUE] [arg FIRST] [arg COUNT]]
[call [cmd "table insert"] [arg TABLE] [arg ROWVALUES] [arg FIRST]]
In the first form, inserts [arg COUNT] rows with value [arg ROWVALUE]
at position [arg FIRST] and returns the new table. In the second form,
[arg ROWVALUES] may be a list of row values or a compatible table.
The rows are inserted at the specified position.
In both cases, the command may extend the array if necessary.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to match
the order of columns to the supplied data. Note that [arg COLUMNS] must
include all columns in the table as the command would not know what
values to use for the unspecified columns.

[call [cmd "table place"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLE] [arg ROWVALUES] [arg INDICES]]
Returns a table with the specified values at the 
corresponding  indices.
[arg ROWVALUES] may be a list of row values or a compatible table. 
The number of rows in [arg ROWVALUES] must not
be less than the number of indices specified in [arg INDICES] and the width
of each row must be no smaller than the width of the table.
[arg INDICES] must be a 
index list in the one of the forms described in [sectref Indices]
and may extend the column if the condition listed there are satisfied.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to target specific columns of the table or to match
the order of columns to the supplied data.

[call [cmd "table put"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLE] [arg ROWVALUES] [opt [arg FIRST]]]
Returns a table with the elements starting at index [arg FIRST]
replaced by the corresponding elements of [arg ROWVALUES].
[arg ROWVALUES] may be a list of values or a table of the same type.
The command may extend the array if necessary. If [arg FIRST] is
not specified the elements are appended.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to target specific columns of the table or to match
the order of columns to the supplied data.

[call [cmd "table range"] [arg [opt "OPTIONS"]] [arg TABLE] [arg LOW] [arg [opt HIGH]]]
Returns all values from a table between the indices specified as a
[sectref Indices range].
Any of the [sectref "Standard Options" "standard options"]
may be specified with this command.

[call [cmd "table reverse"] [arg TABLE]]
Returns the table with order of elements reversed.

[call [cmd "table size"] [arg TABLE]]
Returns the number of rows in the table.

[call [cmd "table"] [cmd "slice"] [arg TABLE] [arg COLUMNLIST]]
Returns a table containing only the specified columns from [arg TABLE].
The columns are specified by their positions or names as a list. A column
must not be included more than once. The returned table contains
columns in the same order as [arg COLUMNLIST].

[call [cmd "table vdelete"] [arg TABLEVAR] [arg INDEX]]
[call [cmd "table vdelete"] [arg TABLEVAR] [arg INDEXLIST]]
[call [cmd "table vdelete"] [arg TABLEVAR] [arg LOW] [arg HIGH]]
Deletes rows from the table in variable [arg TABLEVAR].
The new value is assigned back to the variable. The resulting value
of the variable (which may differ because of traces) is returned as the
result of the command.
Indices are specified in any of the forms described in [sectref Indices].

[call [cmd "table vfill"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLEVAR] [arg ROWVALUE] [arg INDEX]]
[call [cmd "table vfill"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLEVAR] [arg ROWVALUE] [arg INDEXLIST]]
[call [cmd "table vfill"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLEVAR] [arg ROWVALUE] [arg LOW] [arg HIGH]]
Set the elements of the table in variable [arg TABLEVAR] at the specified
indices to
[arg ROWVALUE]. 
The new value is assigned back to the variable. The resulting value
of the variable (which may differ because of traces) is returned as the
result of the command.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to target specific columns of the table or to match
the order of columns to the supplied data.
[para]
See [cmd fill] for more information.

[call [cmd "table vinsert"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLEVAR] [arg ROWVALUE] [arg FIRST] [arg COUNT]]
[call [cmd "table vinsert"] [arg TABLEVAR] [arg ROWVALUES] [arg FIRST]]
In the first form, inserts [arg COUNT] rows with value [arg ROWVALUE]
at position [arg FIRST] in the table stored in variable
[arg TABLEVAR].
In the second form,
[arg ROWVALUES] may be a list of row values or a compatible table.
The rows are inserted at the specified position.
In both cases,
the new value is assigned back to the variable. The resulting value
of the variable (which may differ because of traces) is returned as the
result of the command.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to match
the order of columns to the supplied data. Note that [arg COLUMNS] must
include all columns in the table as the command would not know what
values to use for the unspecified columns.


[call [cmd "table vplace"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLEVAR] [arg ROWVALUES] [arg INDICES]]
Modifies a table stored in the variable [arg TABLEVAR]
with the specified values at the corresponding indices.
The new value is assigned back to the variable. The resulting value
of the variable (which may differ because of traces) is returned as the
result of the command.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to target specific columns of the table or to match
the order of columns to the supplied data.

[para]
See the command [cmd "table place"] for other details.

[call [cmd "table vput"] [opt "[cmd -columns] [arg COLUMNS]"] [arg TABLEVAR] [arg ROWVALUES] [arg FIRST]]
Modifies a variable [arg TABLEVAR] containing a typed column.
The elements of the column starting at index [arg FIRST] are
replaced by the corresponding elements of [arg ROWVALUES]. If [arg FIRST] is
not specified the elements are appended to the array.
The new value is assigned back to the variable. The resulting value
of the variable (which may differ because of traces) is returned as the
result of the command.
[para]
The standard option [sectref "Standard Options" [cmd -columns]] 
may be specified to target specific columns of the table or to match
the order of columns to the supplied data.
[para]
See the command [cmd "table put"] for other details.

[call [cmd "table vreverse"] [arg TABLEVAR]]
Reverses the order of elements in the 
table in variable [arg TABLEVAR], stores it back
in the variable. The result of the command is the resulting value
stored in the variable.

[call [cmd "table width"] [arg TABLE]]
Returns the number of columns in the table.

[list_end]

[manpage_end]
